@@ -36,6 +36,7 @@
 #include <glib.h>
 #include <gst/gst.h>
 #include <gst/base/gstadapter.h>
+#include <gst/controller/gstcontroller.h>
 #include <math.h>
 #include <string.h>
 
@@ -54,6 +55,11 @@
 
 #define DEFAULT_SNR_THRESH 5.5
 
+
+#define GST_CAT_DEFAULT gstlal_itac_debug
+GST_DEBUG_CATEGORY_STATIC(GST_CAT_DEFAULT);
+
+
 static unsigned autocorrelation_channels(const GSTLALItac *element)
 {
 	return gstlal_autocorrelation_chi2_autocorrelation_channels(element->autocorrelation_matrix);
@@ -133,7 +139,7 @@ static GstFlowReturn process(GSTLALItac *element);
 static gboolean sink_event(GstPad *pad, GstEvent *event)
 {
 	GSTLALItac *element = GSTLAL_ITAC(GST_PAD_PARENT(pad));
-	gboolean success;
+	gboolean success = FALSE;
 	GstFlowReturn result;
 
 	switch(GST_EVENT_TYPE(event)) {
@@ -265,7 +271,6 @@ static void set_property(GObject *object, enum property id, const GValue *value,
 	}
 
 	case ARG_AUTOCORRELATION_MASK: {
-		unsigned channels;
 		g_mutex_lock(element->bank_lock);
 
 		if(element->autocorrelation_mask)
@@ -568,8 +573,13 @@ static GstFlowReturn chain(GstPad *pad, GstBuffer *sinkbuf)
 {
 	GSTLALItac *element = GSTLAL_ITAC(gst_pad_get_parent(pad));
 	GstFlowReturn result = GST_FLOW_OK;
+	guint64 maxsize;
+
+	/* do this before accessing any element properties */
+	gst_object_sync_values(G_OBJECT(element), GST_BUFFER_TIMESTAMP(sinkbuf));
+
 	/* The max size to copy from an adapter is the typical output size plus the padding */
-	guint64 maxsize = output_num_bytes(element) + element->adapter->unit_size * element->maxdata->pad * 2;
+	maxsize = output_num_bytes(element) + element->adapter->unit_size * element->maxdata->pad * 2;
 
 	/* if we haven't allocated storage do it now, we should never try to copy from an adapter with a larger buffer than this */
 	if (!element->data)
@@ -648,12 +658,12 @@ static void finalize(GObject *object)
 {
 
 	GSTLALItac *element = GSTLAL_ITAC(object);
-	//FIXME make sure everything is freed
 	g_mutex_free(element->bank_lock);
 	gst_object_unref(element->sinkpad);
 	element->sinkpad = NULL;
 	gst_object_unref(element->srcpad);
 	element->srcpad = NULL;
+	gst_audioadapter_clear(element->adapter);
 	g_object_unref(element->adapter);
 	if (element->instrument) {
 		free(element->instrument);
@@ -666,7 +676,7 @@ static void finalize(GObject *object)
 	if (element->bankarray)
 		free_bank(element);
 	if (element->maxdata) {
-		free(element->maxdata);
+		gstlal_peak_state_free(element->maxdata);
 		element->maxdata = NULL;
 		}
 	if (element->data) {
@@ -722,6 +732,8 @@ static void base_init(gpointer class)
 		"Chad Hanna <chad.hanna@ligo.org>"
 	);
 
+	GST_DEBUG_CATEGORY_INIT(GST_CAT_DEFAULT, "lal_itac", 0, "lal_itac debug category");
+
 	gst_element_class_add_pad_template(
 		element_class,
 		gst_pad_template_new(
@@ -803,7 +815,7 @@ static void class_init(gpointer class, gpointer class_data)
 		g_param_spec_value_array(
 			"sigmasq",
 			"\\sigma^{2} factors",
-			"Vector of \\sigma^{2} factors.",
+			"Vector of \\sigma^{2} factors.  The effective distance of a trigger is \\sqrt{sigma^{2}} / SNR.",
 			g_param_spec_double(
 				"sigmasq",
 				"\\sigma^{2}",
@@ -811,7 +823,7 @@ static void class_init(gpointer class, gpointer class_data)
 				-G_MAXDOUBLE, G_MAXDOUBLE, 0.0,
 				G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS
 			),
-			G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS
+			G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | GST_PARAM_CONTROLLABLE
 		)
 	);
 	
