@@ -26,6 +26,8 @@
 
 
 from collections import deque
+import copy
+import math
 import numpy
 import os
 import resource
@@ -41,6 +43,7 @@ import sys
 import threading
 import time
 import httplib
+import tempfile
 
 # The following snippet is taken from http://gstreamer.freedesktop.org/wiki/FAQ#Mypygstprogramismysteriouslycoredumping.2Chowtofixthis.3F
 import pygtk
@@ -69,8 +72,9 @@ from glue.ligolw.utils import ligolw_sqlite
 from glue.ligolw.utils import ligolw_add
 from glue.ligolw.utils import process as ligolw_process
 from glue.ligolw.utils import search_summary as ligolw_search_summary
+from glue.ligolw.utils import segments as ligolw_segments
+from glue.ligolw.utils import time_slide as ligolw_time_slide
 from pylal import datatypes as laltypes
-from pylal import ligolw_tisi
 from pylal import rate
 from pylal.datatypes import LIGOTimeGPS
 from pylal.datatypes import REAL8FrequencySeries
@@ -173,7 +177,26 @@ def state_vector_on_off_list_from_bits_dict(bit_dict):
 	return onstr, offstr
 
 
-def parse_banks(bank_string):
+def parse_svdbank_string(bank_string):
+	"""
+	parses strings of form 
+	
+	H1:bank1.xml,H2:bank2.xml,L1:bank3.xml
+	
+	into a dictionary of lists of bank files.
+	"""
+	out = {}
+	if bank_string is None:
+		return out
+	for b in bank_string.split(','):
+		ifo, bank = b.split(':')
+		if ifo in out:
+			raise ValueError("Only one svd bank per instrument should be given")
+		out[ifo] = bank
+	return out
+
+
+def parse_iirbank_string(bank_string):
 	"""
 	parses strings of form 
 	
@@ -198,20 +221,26 @@ def parse_bank_files(svd_banks, verbose, snr_threshold = None):
 
 	banks = {}
 
-	for instrument, files in svd_banks.items():
-		for n, filename in enumerate(files):
-			# FIXME over ride the file name stored in the bank file with
-			# this file name this bank I/O code needs to be fixed
-			bank = svd_bank.read_bank(filename, contenthandler = XMLContentHandler, verbose = verbose)
-			bank.template_bank_filename = filename
-			bank.logname = "%sbank%d" % (instrument,n)
-			banks.setdefault(instrument,[]).append(bank)
+	for instrument, filename in svd_banks.items():
+		for n, bank in enumerate(svd_bank.read_banks(filename, contenthandler = LIGOLWContentHandler, verbose = verbose)):
+			# Write out sngl inspiral table to temp file for trigger generator
+			# FIXME teach the trigger generator to get this information a better way
+			bank.template_bank_filename = tempfile.NamedTemporaryFile(suffix = ".gz", delete = False).name
+			xmldoc = ligolw.Document()
+			xmldoc.appendChild(ligolw.LIGO_LW()).appendChild(bank.sngl_inspiral_table)
+			ligolw_utils.write_filename(xmldoc, bank.template_bank_filename, gz = True, verbose = verbose)
+			xmldoc.unlink()	# help garbage collector
+			bank.logname = "%sbank%d" % (instrument, n)
+			banks.setdefault(instrument, []).append(bank)
 			if snr_threshold is not None:
 				bank.snr_threshold = snr_threshold
 
+	# FIXME remove when this is no longer an issue
+	if not banks:
+		raise ValueError("Could not parse bank files into valid bank dictionary.\n\t- Perhaps you are using out-of-date svd bank files?  Please ensure that they were generated with the same code version as the parsing code")
 	return banks
 
-def parse_iirbank_files(iir_banks, verbose, snr_threshold = 5.5):
+def parse_iirbank_files(iir_banks, verbose, snr_threshold = 4.0):
 	"""
 	given a dictionary of lists of iir template bank file names parse them
 	into a dictionary of bank classes
@@ -223,7 +252,7 @@ def parse_iirbank_files(iir_banks, verbose, snr_threshold = 5.5):
 		for n, filename in enumerate(files):
 			# FIXME over ride the file name stored in the bank file with
 			# this file name this bank I/O code needs to be fixed
-			bank = cbc_template_iir.load_iirbank(filename, snr_threshold, contenthandler = XMLContentHandler, verbose = verbose)
+			bank = cbc_template_iir.load_iirbank(filename, snr_threshold, contenthandler = LIGOLWContentHandler, verbose = verbose)
 			bank.template_bank_filename = filename
 			bank.logname = "%sbank%d" % (instrument,n)
 			banks.setdefault(instrument,[]).append(bank)
@@ -231,86 +260,6 @@ def parse_iirbank_files(iir_banks, verbose, snr_threshold = 5.5):
 	return banks
 
 
-
-#
-# add metadata to an xml document in the style of lalapps_inspiral
-#
-
-
-def add_cbc_metadata(xmldoc, process, seg_in):
-	"""
-	A convenience function to add metadata to a cbc output document
-	"""
-	
-	#
-	# add entry to search_summary table
-	#
-
-	try:
-		tbl = lsctables.table.get_table(xmldoc, lsctables.SearchSummaryTable.tableName)
-	except ValueError:
-		tbl = xmldoc.childNodes[-1].appendChild(lsctables.New(lsctables.SearchSummaryTable))
-	search_summary = tbl.RowType()
-	search_summary.process_id = process.process_id
-	search_summary.shared_object = None # FIXME
-	search_summary.lalwrapper_cvs_tag = None # FIXME
-	search_summary.lal_cvs_tag = None # FIXME
-	search_summary.comment = process.comment
-	search_summary.set_ifos(process.get_ifos())
-	search_summary.set_in(seg_in)
-	search_summary.set_out(segments.segment(None, None))
-	search_summary.nevents = None # FIXME
-	search_summary.nnodes = 1
-	tbl.append(search_summary)
-
-	#
-	# add entry to filter table
-	#
-
-	try:
-		tbl = lsctables.table.get_table(xmldoc, lsctables.FilterTable.tableName)
-	except ValueError:
-		tbl = xmldoc.childNodes[-1].appendChild(lsctables.New(lsctables.FilterTable))
-	tbl.sync_next_id()
-	row = tbl.RowType()
-	row.process_id = process.process_id
-	row.program = process.program
-	row.start_time = int(seg_in[0])
-	row.filter_name = None # FIXME
-	row.filter_id = tbl.get_next_id()
-	row.param_set = None # FIXME
-	row.comment = process.comment
-	tbl.append(row)
-
-	#
-	# add entries to search_summvars table
-	#
-
-	try:
-		tbl = lsctables.table.get_table(xmldoc, lsctables.SearchSummVarsTable.tableName)
-	except ValueError:
-		tbl = xmldoc.childNodes[-1].appendChild(lsctables.New(lsctables.SearchSummVarsTable))
-	tbl.sync_next_id()
-	# FIXME
-
-	#
-	# add entries to summ_value table
-	#
-
-	try:
-		tbl = lsctables.table.get_table(xmldoc, lsctables.SummValueTable.tableName)
-	except ValueError:
-		tbl = xmldoc.childNodes[-1].appendChild(lsctables.New(lsctables.SummValueTable))
-	tbl.sync_next_id()
-	# FIXME
-
-	#
-	# done
-	#
-
-	return search_summary
-
-
 #
 # =============================================================================
 #
@@ -320,11 +269,11 @@ def add_cbc_metadata(xmldoc, process, seg_in):
 #
 
 
-class XMLContentHandler(ligolw.LIGOLWContentHandler):
+class LIGOLWContentHandler(ligolw.LIGOLWContentHandler):
 	pass
-ligolw_array.use_in(XMLContentHandler)
-ligolw_param.use_in(XMLContentHandler)
-lsctables.use_in(XMLContentHandler)
+ligolw_array.use_in(LIGOLWContentHandler)
+ligolw_param.use_in(LIGOLWContentHandler)
+lsctables.use_in(LIGOLWContentHandler)
 
 
 #
@@ -375,23 +324,9 @@ def chisq_distribution(df, non_centralities, size):
 #
 
 
-def gen_likelihood_control_doc(far, instruments, name = u"gstlal_inspiral_likelihood", comment = u""):
-	xmldoc = ligolw.Document()
-	node = xmldoc.appendChild(ligolw.LIGO_LW())
-
-	node.appendChild(lsctables.New(lsctables.ProcessTable))
-	node.appendChild(lsctables.New(lsctables.ProcessParamsTable))
-	node.appendChild(lsctables.New(lsctables.SearchSummaryTable))
-	process = ligolw_process.append_process(xmldoc, comment = comment)
-	ligolw_search_summary.append_search_summary(xmldoc, process, ifos = instruments, inseg = far.livetime_seg, outseg = far.livetime_seg)
-
-	node.appendChild(far.to_xml(process, name))
-
-	ligolw_process.set_process_end_time(process)
-	return xmldoc
-
-
 class CoincsDocument(object):
+	sngl_inspiral_columns = ("process_id", "ifo", "end_time", "end_time_ns", "eff_distance", "coa_phase", "mass1", "mass2", "snr", "chisq", "chisq_dof", "bank_chisq", "bank_chisq_dof", "sigmasq", "spin1x", "spin1y", "spin1z", "spin2x", "spin2y", "spin2z", "event_id")
+
 	def __init__(self, filename, process_params, comment, instruments, seg, injection_filename = None, time_slide_file = None, tmp_path = None, replace_file = None, verbose = False):
 		#
 		# how to make another like us
@@ -411,22 +346,28 @@ class CoincsDocument(object):
 
 		self.xmldoc = ligolw.Document()
 		self.xmldoc.appendChild(ligolw.LIGO_LW())
-		self.process = ligolw_process.register_to_xmldoc(self.xmldoc, "gstlal_inspiral", process_params, comment = comment, ifos = instruments)
-		self.search_summary = add_cbc_metadata(self.xmldoc, self.process, seg)
-		# FIXME:  argh, ugly
-		self.xmldoc.childNodes[-1].appendChild(lsctables.New(lsctables.SnglInspiralTable, columns = ("process_id", "ifo", "search", "channel", "end_time", "end_time_ns", "end_time_gmst", "impulse_time", "impulse_time_ns", "template_duration", "event_duration", "amplitude", "eff_distance", "coa_phase", "mass1", "mass2", "mchirp", "mtotal", "eta", "kappa", "chi", "tau0", "tau2", "tau3", "tau4", "tau5", "ttotal", "psi0", "psi3", "alpha", "alpha1", "alpha2", "alpha3", "alpha4", "alpha5", "alpha6", "beta", "f_final", "snr", "chisq", "chisq_dof", "bank_chisq", "bank_chisq_dof", "cont_chisq", "cont_chisq_dof", "sigmasq", "rsqveto_duration", "Gamma0", "Gamma1", "Gamma2", "Gamma3", "Gamma4", "Gamma5", "Gamma6", "Gamma7", "Gamma8", "Gamma9", "spin1x", "spin1y", "spin1z", "spin2x", "spin2y", "spin2z", "event_id")))
+		self.process = ligolw_process.register_to_xmldoc(self.xmldoc, u"gstlal_inspiral", process_params, comment = comment, ifos = instruments)
+		self.search_summary = ligolw_search_summary.append_search_summary(self.xmldoc, self.process,
+			lalwrapper_cvs_tag = None,	# FIXME
+			lal_cvs_tag = None,	# FIXME
+			inseg = seg
+		)
+		self.xmldoc.childNodes[-1].appendChild(lsctables.New(lsctables.SnglInspiralTable, columns = self.sngl_inspiral_columns))
 		self.xmldoc.childNodes[-1].appendChild(lsctables.New(lsctables.CoincDefTable))
 		self.xmldoc.childNodes[-1].appendChild(lsctables.New(lsctables.CoincTable))
 		self.xmldoc.childNodes[-1].appendChild(lsctables.New(lsctables.CoincMapTable))
 		self.xmldoc.childNodes[-1].appendChild(lsctables.New(lsctables.TimeSlideTable))
 		self.xmldoc.childNodes[-1].appendChild(lsctables.New(lsctables.CoincInspiralTable))
+		self.xmldoc.childNodes[-1].appendChild(lsctables.New(lsctables.SegmentDefTable, columns = ligolw_segments.LigolwSegmentList.segment_def_columns))
+		self.xmldoc.childNodes[-1].appendChild(lsctables.New(lsctables.SegmentSumTable, columns = ligolw_segments.LigolwSegmentList.segment_sum_columns))
+		self.xmldoc.childNodes[-1].appendChild(lsctables.New(lsctables.SegmentTable, columns = ligolw_segments.LigolwSegmentList.segment_columns))
 
 		#
 		# optionally insert injection list document
 		#
 
 		if injection_filename is not None:
-			ligolw_add.ligolw_add(self.xmldoc, [injection_filename], contenthandler = XMLContentHandler, verbose = verbose)
+			ligolw_add.ligolw_add(self.xmldoc, [injection_filename], contenthandler = LIGOLWContentHandler, verbose = verbose)
 
 		#
 		# optionally insert a time slide table document.  if we
@@ -434,13 +375,12 @@ class CoincsDocument(object):
 		# duplicate offset vectors when done
 		#
 
-		time_slide_table = lsctables.table.get_table(self.xmldoc, lsctables.TimeSlideTable.tableName)
+		time_slide_table = lsctables.TimeSlideTable.get_table(self.xmldoc)
 		if time_slide_file is not None:
-			ligolw_add.ligolw_add(self.xmldoc, [time_slide_file], contenthandler = XMLContentHandler, verbose = verbose)
+			ligolw_add.ligolw_add(self.xmldoc, [time_slide_file], contenthandler = LIGOLWContentHandler, verbose = verbose)
 		else:
-			for row in ligolw_tisi.RowsFromOffsetDict(dict.fromkeys(instruments, 0.0), time_slide_table.get_next_id(), self.process):
-				time_slide_table.append(row)
-		time_slide_mapping = ligolw_tisi.time_slides_vacuum(time_slide_table.as_dict())
+			time_slide_table.append_offsetvector(dict.fromkeys(instruments, 0.0), self.process)
+		time_slide_mapping = ligolw_time_slide.time_slides_vacuum(time_slide_table.as_dict())
 		iterutils.inplace_filter(lambda row: row.time_slide_id not in time_slide_mapping, time_slide_table)
 		for tbl in self.xmldoc.getElementsByTagName(ligolw.Table.tagName):
 			tbl.applyKeyMapping(time_slide_mapping)
@@ -452,18 +392,9 @@ class CoincsDocument(object):
 		#
 
 		if filename is not None and filename.endswith('.sqlite'):
-			# FIXME:  remove the ID remap stuff when we can
-			# rely on having glue 1.44
 			self.working_filename = dbtables.get_connection_filename(filename, tmp_path = tmp_path, replace_file = replace_file, verbose = verbose)
 			self.connection = sqlite3.connect(self.working_filename, check_same_thread = False)
-			dbtables.idmap_create(self.connection)
-			dbtables.idmap_sync(self.connection)
-			__orig_append, dbtables.DBTable.append = dbtables.DBTable.append, dbtables.DBTable._remapping_append
 			ligolw_sqlite.insert_from_xmldoc(self.connection, self.xmldoc, preserve_ids = False, verbose = verbose)
-			dbtables.DBTable.append = __orig_append
-			del __orig_append
-			dbtables.idmap_reset(self.connection)
-			dbtables.idmap_sync(self.connection)
 
 			#
 			# convert self.xmldoc into wrapper interface to
@@ -489,7 +420,8 @@ class CoincsDocument(object):
 		# know if they are database-backed or XML objects
 		#
 
-		self.sngl_inspiral_table = lsctables.table.get_table(self.xmldoc, lsctables.SnglInspiralTable.tableName)
+		self.sngl_inspiral_table = lsctables.SnglInspiralTable.get_table(self.xmldoc)
+		self.llwsegments = ligolw_segments.LigolwSegments(self.xmldoc, self.process)
 
 
 	def commit(self):
@@ -523,14 +455,14 @@ class CoincsDocument(object):
 
 	def T050017_filename(self, description, extension):
 		start, end = self.search_summary_outseg
-		return "%s-%s-%d-%d.%s" % ("".join(sorted(self.process.get_ifos())), description, int(start), int(end - start), extension)
+		start, end = int(math.floor(start)), int(math.ceil(end))
+		return "%s-%s-%d-%d.%s" % ("".join(sorted(self.process.get_ifos())), description, start, end - start, extension)
 
 
 	def write_output_file(self, verbose = False):
+		self.llwsegments.finalize()
 		ligolw_process.set_process_end_time(self.process)
 
-		# FIXME:  should signal trapping be disabled in this code
-		# path?  I think not
 		if self.connection is not None:
 			seg = self.search_summary_outseg
 			# record the final state of the search_summary and
@@ -544,8 +476,8 @@ class CoincsDocument(object):
 			self.connection.commit()
 			dbtables.build_indexes(self.connection, verbose = verbose)
 			self.connection.close()
-			dbtables.put_connection_filename(self.filename, self.working_filename, verbose = verbose)
 			self.connection = None
+			dbtables.put_connection_filename(self.filename, self.working_filename, verbose = verbose)
 		else:
 			self.sngl_inspiral_table.sort(lambda a, b: cmp(a.end_time, b.end_time) or cmp(a.end_time_ns, b.end_time_ns) or cmp(a.ifo, b.ifo))
 			self.search_summary.nevents = len(self.sngl_inspiral_table)
@@ -553,12 +485,29 @@ class CoincsDocument(object):
 
 
 class Data(object):
-	def __init__(self, filename, process_params, pipeline, instruments, seg, coincidence_threshold, FAR, marginalized_likelihood_file = None, injection_filename = None, time_slide_file = None, comment = None, tmp_path = None, assign_likelihoods = False, likelihood_snapshot_interval = None, thinca_interval = 50.0, gracedb_far_threshold = None, sngls_snr_threshold = 8.0, likelihood_file = None, gracedb_group = "Test", gracedb_type = "LowMass", replace_file = True, verbose = False):
+	def __init__(self, filename, process_params, pipeline, instruments, seg, coincidence_threshold, coinc_params_distributions, marginalized_likelihood_file = None, injection_filename = None, time_slide_file = None, comment = None, tmp_path = None, likelihood_snapshot_interval = None, thinca_interval = 50.0, sngls_snr_threshold = None, gracedb_far_threshold = None, gracedb_group = "Test", gracedb_type = "LowMass", replace_file = True, verbose = False):
 		#
 		# initialize
 		#
-		
+
+		self.lock = threading.Lock()
+		self.pipeline = pipeline
+		self.verbose = verbose
+		# None to disable likelihood ratio assignment, otherwise a filename
+		self.marginalized_likelihood_file = marginalized_likelihood_file
+		# None to disable periodic snapshots, otherwise seconds
+		# set to 1.0 to disable background data decay
+		self.likelihood_snapshot_interval = likelihood_snapshot_interval
+		self.likelihood_snapshot_timestamp = None
+		# gracedb far threshold
+		self.gracedb_far_threshold = gracedb_far_threshold
+		self.gracedb_group = gracedb_group
+		self.gracedb_type = gracedb_type
+
+		#
 		# setup bottle routes
+		#
+
 		bottle.route("/latency_histogram.txt")(self.web_get_latency_histogram)
 		bottle.route("/latency_history.txt")(self.web_get_latency_history)
 		bottle.route("/snr_history.txt")(self.web_get_snr_history)
@@ -569,39 +518,12 @@ class Data(object):
 		bottle.route("/sngls_snr_threshold.txt", method = "GET")(self.web_get_sngls_snr_threshold)
 		bottle.route("/sngls_snr_threshold.txt", method = "POST")(self.web_set_sngls_snr_threshold)
 
-		self.lock = threading.Lock()
-		self.pipeline = pipeline
-		self.instruments = instruments
-		self.verbose = verbose
-		# True to enable likelihood assignment
-		self.assign_likelihoods = assign_likelihoods
-		self.marginalized_likelihood_file = marginalized_likelihood_file
-		# Set to None to disable period snapshots, otherwise set to seconds
-		self.likelihood_snapshot_interval = likelihood_snapshot_interval
-		# Set to 1.0 to disable background data decay
-		# FIXME:  should this live in the DistributionsStats object?
-		self.likelihood_snapshot_timestamp = None
-		# gracedb far threshold
-		self.gracedb_far_threshold = gracedb_far_threshold
-		self.gracedb_group = gracedb_group
-		self.gracedb_type = gracedb_type
-
 		#
 		# initialize document to hold coincs and segments
 		#
 
 		self.coincs_document = CoincsDocument(filename, process_params, comment, instruments, seg, injection_filename = injection_filename, time_slide_file = time_slide_file, tmp_path = tmp_path, replace_file = replace_file, verbose = verbose)
 
-		#
-		# setup far/fap book-keeping
-		#
-
-		self.far = FAR
-		self.ranking_data = None
-		if self.assign_likelihoods:
-			self.far.smooth_distribution_stats(verbose = verbose)
-		self.likelihood_file = likelihood_file
-
 		#
 		# attach a StreamThinca instance to ourselves
 		#
@@ -609,10 +531,18 @@ class Data(object):
 		self.stream_thinca = streamthinca.StreamThinca(
 			coincidence_threshold = coincidence_threshold,
 			thinca_interval = thinca_interval,	# seconds
-			trials_table = self.far.trials_table,
 			sngls_snr_threshold = sngls_snr_threshold
 		)
 
+		#
+		# setup likelihood ratio book-keeping.  seglistsdicts to be
+		# populated the pipeline handler
+		#
+
+		self.coinc_params_distributions = coinc_params_distributions
+		self.seglistdicts = None
+		self.fapfar = None
+
 		#
 		# Fun output stuff
 		#
@@ -627,22 +557,15 @@ class Data(object):
 			# retrieve triggers from appsink element
 			buf = elem.emit("pull-buffer")
 			events = sngl_inspirals_from_buffer(buf)
+			# FIXME:  ugly way to get the instrument
+			instrument = elem.get_name().split("_")[0]
 
-			# update search_summary out segment.  note that
-			# both the trigger document and the FAR object get
-			# their own copies of the segment.  the segment in
-			# self.search_summary is what gets recorded in the
-			# trigger document, the segment in the FAR object
-			# gets used for online FAR/FAP assignment and is
-			# what gets recorded in the likelihood data
-			# document.
+			# update search_summary out segment and our
+			# livetime
 			buf_timestamp = LIGOTimeGPS(0, buf.timestamp)
-			buf_end_time = buf_timestamp + LIGOTimeGPS(0, buf.duration)
-			self.coincs_document.add_to_search_summary_outseg(segments.segment(buf_timestamp, buf_end_time))
-			if self.far.livetime_seg == segments.segment(None, None):
-				self.far.livetime_seg = self.coincs_document.search_summary_outseg
-			else:
-				self.far.livetime_seg = segments.segmentlist([self.coincs_document.search_summary_outseg, self.far.livetime_seg]).extent()
+			buf_seg = segments.segment(buf_timestamp, buf_timestamp + LIGOTimeGPS(0, buf.duration))
+			self.coincs_document.add_to_search_summary_outseg(buf_seg)
+			self.seglistdicts["triggersegments"][instrument] |= segments.segmentlist((buf_seg,))
 
 			# set metadata on triggers.  because this uses the
 			# ID generator attached to the database-backed
@@ -655,8 +578,13 @@ class Data(object):
 				event.event_id = self.coincs_document.get_next_sngl_id()
 
 			# update likelihood snapshot if needed
-			if self.likelihood_snapshot_interval is not None and (self.likelihood_snapshot_timestamp is None or (self.likelihood_snapshot_interval is not None and buf_timestamp - self.likelihood_snapshot_timestamp >= self.likelihood_snapshot_interval)):
+			if self.likelihood_snapshot_interval is not None and (self.likelihood_snapshot_timestamp is None or buf_timestamp - self.likelihood_snapshot_timestamp >= self.likelihood_snapshot_interval):
 				self.likelihood_snapshot_timestamp = buf_timestamp
+
+				# smooth the distributions.  re-populates
+				# PDF arrays from raw counts
+				self.coinc_params_distributions.finish(verbose = self.verbose)
+
 				# post a checkpoint message.  FIXME:  make
 				# sure this triggers
 				# self.snapshot_output_file() to be
@@ -666,69 +594,123 @@ class Data(object):
 				# code should be responsible for it
 				# somehow, no?
 				self.pipeline.get_bus().post(message_new_checkpoint(self.pipeline, timestamp = buf_timestamp.ns()))
-				if self.assign_likelihoods:
-					assert self.marginalized_likelihood_file is not None
-					# smooth the distribution_stats
-					self.far.smooth_distribution_stats(verbose = self.verbose)
-					# update stream thinca's likelihood data
-					self.stream_thinca.set_likelihood_data(self.far.distribution_stats.smoothed_distributions, self.far.distribution_stats.likelihood_params_func)
 
-					# Read in the the background likelihood distributions that should have been updated asynchronously
-					self.ranking_data, procid = far.RankingData.from_xml(ligolw_utils.load_filename(self.marginalized_likelihood_file, verbose = self.verbose, contenthandler = XMLContentHandler))
-					self.ranking_data.compute_joint_cdfs()
+				if self.marginalized_likelihood_file is not None:
+					# FIXME:  must set horizon
+					# distances in coinc params object
 
-					# set up the scale factor for the trials table to normalize the rate
-					for ifos in self.ranking_data.scale:
-						try:
-							self.ranking_data.scale[ifos] = (self.ranking_data.trials_table[ifos].count_below_thresh or 1) / self.ranking_data.trials_table[ifos].thresh / float(abs(self.ranking_data.livetime_seg)) * (self.ranking_data.trials_table.num_nonzero_count() or 1)
-						except TypeError:
-							self.ranking_data.scale[ifos] = 1
-							print >>sys.stderr, "could not set scale factor, probably because we do not have live time info yet.  seg is: %s" % str(self.ranking_data.livetime_seg)
+					# enable streamthinca's likelihood
+					# ratio assignment using our own,
+					# local, parameter distribution
+					# data
+					self.stream_thinca.coinc_params_distributions = self.coinc_params_distributions
 
-					# write the new distribution stats to disk
-					ligolw_utils.write_filename(gen_likelihood_control_doc(self.far, self.instruments), self.likelihood_file, gz = (self.likelihood_file or "stdout").endswith(".gz"), verbose = False, trap_signals = None)
-				else:
-					self.ranking_data = None
+					# read the marginalized likelihood
+					# ratio distributions that have
+					# been updated asynchronously and
+					# initialize a FAP/FAR assignment
+					# machine from it.  NOTE:  to keep
+					# overhead low we do not .finish()
+					# the ranking data here;  the
+					# external process generating this
+					# input file must do that for us.
+					coinc_params_distributions, ranking_data, seglists = far.parse_likelihood_control_doc(ligolw_utils.load_filename(self.marginalized_likelihood_file, verbose = self.verbose, contenthandler = far.ThincaCoincParamsDistributions.LIGOLWContentHandler))
+					if coinc_params_distributions is None:
+						raise ValueError("\"%s\" does not contain coinc parameter PDFs" % self.marginalized_likelihood_file)
+					if ranking_data is None:
+						raise ValueError("\"%s\" does not contain ranking statistic PDFs" % self.marginalized_likelihood_file)
+					# we're using the class attribute
+					# elsewhere so make sure these two
+					# match
+					assert ranking_data.ln_likelihood_ratio_threshold == far.RankingData.ln_likelihood_ratio_threshold
+					self.fapfar = far.FAPFAR(ranking_data, coinc_params_distributions.count_above_threshold, threshold = far.RankingData.ln_likelihood_ratio_threshold, livetime = far.get_live_time(seglists))
 
-			# run stream thinca
-			noncoinc_sngls = self.stream_thinca.add_events(self.coincs_document.xmldoc, self.coincs_document.process_id, events, buf_timestamp, FAP = self.ranking_data)
-
-			# update the parameter distribution data.  only
-			# update from sngls that weren't used in coincs
-			for event in noncoinc_sngls:
-				self.far.distribution_stats.add_single(event)
-
-			# update output document
+			# run stream thinca.  update the parameter
+			# distribution data from sngls that weren't used in
+			# coincs
+			for event in self.stream_thinca.add_events(self.coincs_document.xmldoc, self.coincs_document.process_id, events, buf_timestamp, fapfar = self.fapfar):
+				self.coinc_params_distributions.add_background(self.coinc_params_distributions.coinc_params((event,), None))
 			self.coincs_document.commit()
 
+			# Cluster last coincs before recording number of zero
+			# lag events or sending alerts to gracedb
+			# FIXME Do proper clustering that saves states between
+			# thinca intervals and uses an independent clustering
+			# window. This can also go wrong if there are multiple
+			# events with an identical likelihood.  It will just
+			# choose the event with the highest event id
+			# FIXME uncomment for clustering
+			#if self.stream_thinca.last_coincs:
+			#	self.stream_thinca.last_coincs.coinc_event_index = dict([max(self.stream_thinca.last_coincs.coinc_event_index.iteritems(), key = lambda (coinc_event_id, coinc_event): coinc_event.likelihood)])
+
+			# update zero-lag coinc bin counts in
+			# coinc_params_distributions.  NOTE:  if likelihood
+			# ratios are known then these are the counts of
+			# occurances of parameters in coincs above
+			# threshold, otherwise they are the counts of
+			# occurances of parameters in all coincs.  knowing
+			# the meaning of the counts that get recorded is
+			# left as an exercise to the user
+			if self.stream_thinca.last_coincs:
+				for coinc_event_id, coinc_event in self.stream_thinca.last_coincs.coinc_event_index.items():
+					offset_vector = self.stream_thinca.last_coincs.offset_vector(coinc_event.time_slide_id)
+					if (coinc_event.likelihood >= far.RankingData.ln_likelihood_ratio_threshold or self.marginalized_likelihood_file is None) and not any(offset_vector.values()):
+						self.coinc_params_distributions.add_zero_lag(self.coinc_params_distributions.coinc_params(self.stream_thinca.last_coincs.sngl_inspirals(coinc_event_id), offset_vector))
+
 			# do GraceDB alerts
 			if self.gracedb_far_threshold is not None:
 				self.__do_gracedb_alerts()
 				self.__update_eye_candy()
 
-	def __web_get_likelihood_file(self):
-		# generate a coinc parameter distribution file
-		output = StringIO.StringIO()
-		ligolw_utils.write_fileobj(gen_likelihood_control_doc(self.far, self.instruments), output, trap_signals = None)
-		outstr = output.getvalue()
-		output.close()
-		return outstr
+	def record_horizon_distance(self, instrument, timestamp, psd, m1, m2, snr_threshold = 8.0):
+		with self.lock:
+			horizon_distance = reference_psd.horizon_distance(psd, m1 = m1, m2 = m2, snr = snr_threshold, f_min = 10.0)
+			# NOTE:  timestamp is cast to float.  should be
+			# safe, whitener should be reporting PSDs with
+			# integer timestamps.  anyway, we don't need
+			# nanosecond precision for the horizon distance
+			# history.
+			try:
+				horizon_history = self.coinc_params_distributions.horizon_history[instrument]
+			except KeyError:
+				horizon_history = self.coinc_params_distributions.horizon_history[instrument] = far.NearestLeafTree()
+			horizon_history[float(timestamp)] = horizon_distance
+
+	def __get_likelihood_file(self):
+		# generate a coinc parameter distribution document.  NOTE:
+		# likelihood ratio PDFs are *not* included.
+		xmldoc = ligolw.Document()
+		xmldoc.appendChild(ligolw.LIGO_LW())
+		process = ligolw_process.register_to_xmldoc(xmldoc, u"gstlal_inspiral", paramdict = {})
+		search_summary = ligolw_search_summary.append_search_summary(xmldoc, process, ifos = self.seglistdicts["triggersegments"].keys(), inseg = self.seglistdicts["triggersegments"].extent_all(), outseg = self.seglistdicts["triggersegments"].extent_all())
+		# FIXME:  now that we've got all kinds of segment lists
+		# being collected, decide which of them should go here.
+		far.gen_likelihood_control_doc(xmldoc, process, self.coinc_params_distributions, None, self.seglistdicts["triggersegments"])
+		ligolw_process.set_process_end_time(process)
+		return xmldoc
 
 	def web_get_likelihood_file(self):
 		with self.lock:
-			return self.__web_get_likelihood_file()
+			output = StringIO.StringIO()
+			ligolw_utils.write_fileobj(self.__get_likelihood_file(), output, trap_signals = None)
+			outstr = output.getvalue()
+			output.close()
+			return outstr
 
 	def __flush(self):
 		# run StreamThinca's .flush().  returns the last remaining
 		# non-coincident sngls.  add them to the distribution
-		if self.assign_likelihoods:
-			FAP = self.ranking_data
-		else:
-			FAP = None
-		for event in self.stream_thinca.flush(self.coincs_document.xmldoc, self.coincs_document.process_id, FAP = FAP):
-			self.far.distribution_stats.add_single(event)
+		for event in self.stream_thinca.flush(self.coincs_document.xmldoc, self.coincs_document.process_id, fapfar = self.fapfar):
+			self.coinc_params_distributions.add_background(self.coinc_params_distributions.coinc_params((event,), None))
 		self.coincs_document.commit()
 
+		# update zero-lag bin counts in coinc_params_distributions
+		if self.stream_thinca.last_coincs:
+			for coinc_event_id, coinc_event in self.stream_thinca.last_coincs.coinc_event_index.items():
+				offset_vector = self.stream_thinca.last_coincs.offset_vector(coinc_event.time_slide_id)
+				if (coinc_event.likelihood >= far.RankingData.ln_likelihood_ratio_threshold or self.marginalized_likelihood_file is None) and not any(offset_vector.values()):
+					self.coinc_params_distributions.add_zero_lag(self.coinc_params_distributions.coinc_params(self.stream_thinca.last_coincs.sngl_inspirals(coinc_event_id), offset_vector))
+
 		# do GraceDB alerts
 		if self.gracedb_far_threshold is not None:
 			self.__do_gracedb_alerts()
@@ -744,21 +726,18 @@ class Data(object):
 			psdmessage = None
 			coinc_inspiral_index = self.stream_thinca.last_coincs.coinc_inspiral_index
 
-			# FIXME:  this is hacked to only send at most the
-			# one best coinc in this set.  May not make sense
-			# depending on what is in last_coincs.  FIX
-			# PROPERLY.  This is probably mostly okay because
-			# we should be doing coincidences every 10s which
-			# is a reasonable time to cluster over.  the slice
-			# can be edited (or removed) to change this.
-			for likelihood, coinc_event_id in sorted((coinc_event.likelihood, coinc_event.coinc_event_id) for coinc_event in self.stream_thinca.last_coincs.coinc_event_index.values())[-1:]:
+			# This appears to be a silly for loop since
+			# coinc_event_index will only have one value, but we're
+			# future proofing this at the point where it could have
+			# multiple clustered events
+			for coinc_event in self.stream_thinca.last_coincs.coinc_event_index.values():
 				#
-				# quit if the false alarm rate is not low
-				# enough, or is nan
+				# continue if the false alarm rate is not low
+				# enough, or is nan.
 				#
 
-				if coinc_inspiral_index[coinc_event_id].combined_far > self.gracedb_far_threshold or numpy.isnan(coinc_inspiral_index[coinc_event_id].combined_far):
-					break
+				if coinc_inspiral_index[coinc_event.coinc_event_id].combined_far > self.gracedb_far_threshold or numpy.isnan(coinc_inspiral_index[coinc_event.coinc_event_id].combined_far):
+					continue
 
 				#
 				# retrieve PSDs
@@ -768,7 +747,7 @@ class Data(object):
 					if self.verbose:
 						print >>sys.stderr, "retrieving PSDs from whiteners and generating psd.xml.gz ..."
 					psddict = {}
-					for instrument in self.instruments:
+					for instrument in self.seglistdicts["triggersegments"]:
 						elem = self.pipeline.get_by_name("lal_whiten_%s" % instrument)
 						psddict[instrument] = REAL8FrequencySeries(
 							name = "PSD",
@@ -785,10 +764,10 @@ class Data(object):
 				# fake a filename for end-user convenience
 				#
 
-				observatories = "".join(sorted(set(instrument[0] for instrument in self.instruments)))
-				instruments = "".join(sorted(self.instruments))
-				description = "%s_%s_%s_%s" % (instruments, ("%.4g" % coinc_inspiral_index[coinc_event_id].mass).replace(".", "_").replace("-", "_"), self.gracedb_group, self.gracedb_type)
-				end_time = int(coinc_inspiral_index[coinc_event_id].get_end())
+				observatories = "".join(sorted(set(instrument[0] for instrument in self.seglistdicts["triggersegments"])))
+				instruments = "".join(sorted(self.seglistdicts["triggersegments"]))
+				description = "%s_%s_%s_%s" % (instruments, ("%.4g" % coinc_inspiral_index[coinc_event.coinc_event_id].mass).replace(".", "_").replace("-", "_"), self.gracedb_group, self.gracedb_type)
+				end_time = int(coinc_inspiral_index[coinc_event.coinc_event_id].get_end())
 				filename = "%s-%s-%d-%d.xml" % (observatories, description, end_time, 0)
 
 				#
@@ -805,7 +784,19 @@ class Data(object):
 				if self.verbose:
 					print >>sys.stderr, "sending %s to gracedb ..." % filename
 				message = StringIO.StringIO()
-				ligolw_utils.write_fileobj(self.stream_thinca.last_coincs[coinc_event_id], message, gz = False, trap_signals = None)
+				xmldoc = self.stream_thinca.last_coincs[coinc_event.coinc_event_id]
+				# give the alert all the standard inspiral
+				# columns (attributes should all be
+				# populated).  FIXME:  ugly.
+				sngl_inspiral_table = lsctables.SnglInspiralTable.get_table(xmldoc)
+				for standard_column in ("process_id", "ifo", "search", "channel", "end_time", "end_time_ns", "end_time_gmst", "impulse_time", "impulse_time_ns", "template_duration", "event_duration", "amplitude", "eff_distance", "coa_phase", "mass1", "mass2", "mchirp", "mtotal", "eta", "kappa", "chi", "tau0", "tau2", "tau3", "tau4", "tau5", "ttotal", "psi0", "psi3", "alpha", "alpha1", "alpha2", "alpha3", "alpha4", "alpha5", "alpha6", "beta", "f_final", "snr", "chisq", "chisq_dof", "bank_chisq", "bank_chisq_dof", "cont_chisq", "cont_chisq_dof", "sigmasq", "rsqveto_duration", "Gamma0", "Gamma1", "Gamma2", "Gamma3", "Gamma4", "Gamma5", "Gamma6", "Gamma7", "Gamma8", "Gamma9", "spin1x", "spin1y", "spin1z", "spin2x", "spin2y", "spin2z", "event_id"):
+					try:
+						sngl_inspiral_table.appendColumn(bonus_column)
+					except ValueError:
+						# already has it
+						pass
+				ligolw_utils.write_fileobj(xmldoc, message, gz = False, trap_signals = None)
+				xmldoc.unlink()
 				# FIXME: make this optional from command line?
 				if True:
 					resp = gracedb_client.createEvent(self.gracedb_group, self.gracedb_type, filename, message.getvalue())
@@ -930,29 +921,44 @@ class Data(object):
 
 	def __write_output_file(self, filename = None, likelihood_file = None, verbose = False):
 		self.__flush()
+
+		# FIXME:  should this be done in .flush() somehow?
+		for segtype, seglistdict in self.seglistdicts.items():
+			self.coincs_document.llwsegments.insert_from_segmentlistdict(seglistdict, name = segtype, comment = "LLOID")
+
 		if filename is not None:
 			self.coincs_document.filename = filename
 		self.coincs_document.write_output_file(verbose = verbose)
 
-		# write out the snr / chisq histograms
+		# write the parameter PDF file.  NOTE;  this file contains
+		# raw bin counts, and might or might not contain smoothed,
+		# normalized, PDF arrays but if it does they will not
+		# necessarily correspond to the bin counts, and it
+		# certainly does not include likelihood ratio PDF data.
+		#
+		# the parameter PDF arrays cannot be re-computed here
+		# because it would interfer with their use by stream
+		# thinca.  we want to know precisely when the arrays get
+		# updated so we can have a hope of computing the likelihood
+		# ratio PDFs correctly.
 		if likelihood_file is None:
-			fname = os.path.split(self.coincs_document.filename)
+			likelihood_file = os.path.split(self.coincs_document.filename)
 			try: # preserve LIGO-T010150-00 if possible 
-				ifo, desc, start, dur = ".".join(fname[1].split('.')[:-1]).split('-')
-				fname = os.path.join(fname[0], '%s-%s_SNR_CHI-%s-%s.xml.gz' % (ifo, desc, start, dur))
+				ifo, desc, start, dur = ".".join(likelihood_file[1].split('.')[:-1]).split('-')
+				likelihood_file = os.path.join(likelihood_file[0], '%s-%s_SNR_CHI-%s-%s.xml.gz' % (ifo, desc, start, dur))
 			except ValueError:
-				fname = os.path.join(fname[0], '%s_SNR_CHI.xml.gz' % fname[1].split('.')[0])
-		else:
-			fname = likelihood_file
-		ligolw_utils.write_filename(gen_likelihood_control_doc(self.far, self.instruments), fname, gz = (fname or "stdout").endswith(".gz"), verbose = verbose, trap_signals = None)
+				likelihood_file = os.path.join(likelihood_file[0], '%s_SNR_CHI.xml.gz' % likelihood_file[1].split('.')[0])
+		ligolw_utils.write_filename(self.__get_likelihood_file(), likelihood_file, gz = (likelihood_file or "stdout").endswith(".gz"), verbose = verbose, trap_signals = None)
+
+		# can't be used anymore
+		del self.coincs_document
 
 	def write_output_file(self, filename = None, likelihood_file = None, verbose = False):
 		with self.lock:
 			self.__write_output_file(filename = filename, likelihood_file = likelihood_file, verbose = verbose)
-			# can't be used anymore
-			del self.coincs_document
 
 	def snapshot_output_file(self, description, extension, verbose = False):
 		with self.lock:
+			coincs_document = self.coincs_document.get_another()
 			self.__write_output_file(filename = self.coincs_document.T050017_filename(description, extension), verbose = verbose)
-			self.coincs_document = self.coincs_document.get_another()
+			self.coincs_document = coincs_document
